/*
    This file is part of WinterHill

    WinterHill is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version. 

    WinterHill is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with WinterHill.  If not, see <https://www.gnu.org/licenses/>.
    
    Copyright Brian Jordan, G4EWJ, 2020
*/

// compiled by: Borland C++ 5.5.1 for Win32 Copyright (c) 1993, 2000 Borland

#define VERSIONX        "3v20"
#define VERSIONX2		"d"


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@
//@    Usage: whpcviewer-3v20  IPADDRESS BASEIPPORT IPINTERFACEADDRESS 
//@                            VLCLEFT VLCTOP VLCWIDTH VLCHEIGHT INFOLEFT INFOTOP INFOWIDTH INFOHEIGHT
//@
//@    whpcviewer starts 4 VLC windows and directs them to 4 transports streams on ports BASEIPPORT+41,42,43,44
//@    the IPADDRESS is either zero or should match the broadcast address used on the Raspberry Pi
//@    the BASEIPPORT should match that used on the Raspberry Pi
//@    coordinate 0,0 is the top left of the screen
//@    the INFO window position is relative to the bottom left of VLC window 3 
//@    the VLC title bar is updated from the status stream on port BASEIPPORT+1  
//@
//@    when the info windeo is selected, commands are:
//@        d       display all windows
//@        h       hide all windows
//@        s       toggle half size display
//@        ENTER   quit
//@        
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@  

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>
#include <string.h>
#include <dos.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <io.h>
#include <errno.h>
#include <dir.h>
#include <signal.h>
#include <sys\timeb.h>
#include <msregdb.h>
#include <winbase.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <iphlpapi>

typedef unsigned int	uint32 ;
typedef signed int		int32 ;
typedef unsigned short	uint16 ;
typedef unsigned char	uint8 ;

#define VERSIONX        "3v20"
#define VERSIONX2		"b"

#define CR				13
#define LF				10
#define ESC				27	
#define BACK			8
#define MAXRECEIVERS	4
#define MAXINFOS		100
#define PORTINFOLMEX	1					// offset from base port for expanded info for all receivers
#define PORTINFOMULTI	2					// offset from base port for 4 line text summary for all receivers

/* define the various status reports in rcv[rx].textinfos - only 0 94 96 97 99 are used*/

#define STATUS_RECEIVER_NUMBER     0
#define STATUS_STATE               1
#define STATUS_LNA_GAIN            2
#define STATUS_PUNCTURE_RATE       3
#define STATUS_POWER_I             4 
#define STATUS_POWER_Q             5
#define STATUS_CARRIER_FREQUENCY   6
#define STATUS_CONSTELLATION_I     7
#define STATUS_CONSTELLATION_Q     8
#define STATUS_SYMBOL_RATE         9
#define STATUS_VITERBI_ERROR_RATE 10
#define STATUS_BER                11
#define STATUS_MER                12
#define STATUS_SERVICE_NAME       13		// usually callsign
#define STATUS_SERVICE_PROVIDER_NAME  14
#define STATUS_TS_NULL_PERCENTAGE 15
#define STATUS_ES_PID             16
#define STATUS_ES_TYPE            17
#define STATUS_MODCOD             18
#define STATUS_FRAME_TYPE         19
#define STATUS_PILOTS             20
#define STATUS_ERRORS_LDPC_COUNT  21
#define STATUS_ERRORS_BCH_COUNT   22
#define STATUS_ERRORS_BCH_UNCORRECTED   23
#define STATUS_LNB_SUPPLY         24
#define STATUS_LNB_POLARISATION_H 25

// new statuses added for WinterHill

#define STATUS_MULTISTREAM0   	  26		// B3:B2:B1:B0 = PDELCTR1,PDELCTRL0,MATSTR1,MATSTR0  
#define STATUS_MULTISTREAM1   	  27		//       B1:B0 = ISIBITENA,ISIENTRY
#define STATUS_DEBUG0			  28		//		 B1:B0 = TSSTATUS2,TSSTATUS 
#define STATUS_DEBUG1			  29		//		 B1:B0 = TSSTATUS,TSSTATUS
#define STATUS_DNUMBER			  30		// MER value - MER decode threshold (can be negative)		
#define STATUS_SERVICE_TYPE		  31		// 0x02=MPEG2, 0x1b=H264, 0x24=H265	
#define STATUS_ROLLOFF			  32		
#define STATUS_ANTENNA            33        // 1:2 = TOP:BOT
#define STATUS_AUDIO_TYPE         34        // 0x03=MP2, 0x0f=AAC

#define STATUS_TITLEBAR           94        // the VLC titlebar generated on the RPi

#define STATUS_VLCSTOPS           96        // STOP commands sent by the WH main application
#define STATUS_VLCNEXTS           97        // NEXT commands sent by the WH main application
#define STATUS_MODECHANGES        98        // incremented for newcommand, callsign, codec
#define STATUS_IPCHANGES          99        // incremented when a command comes from a different IP address

struct receiver
{
			uint32					active ;
			uint32					scanstate ;
			char					textinfos [MAXINFOS][256] ;
			uint32					modechanges ;
            uint32                  lastvlcstops ;
            uint32                  lastvlcnexts ;
} ;

// global variables

			struct receiver			rcv [MAXRECEIVERS+1] ;
            char                    title[] = "WinterHill" ;
volatile    int                     stop ; 
volatile    int                     children = 0 ;
            HWND                    vlcwindowhandle [5] ;
			HWND					infowindowhandle ;
            STARTUPINFO             si ;
            PROCESS_INFORMATION     pi [5] ;
			HWND                	mainwindowhandle ;
			uint16    		        baseport ;
			char		    		address[16] ;					
			char		    		interfaceaddress[16] ;			// network interface to use, if more than one is available
			char					vlcpath [256] ;
			char					infoname [256] ;
			DWORD					myprocessid ;	
			int         	        vlcwidth, vlcheight, vlctop, vlcleft ;
			int    		            infowidth, infoheight, infotop, infoleft ;
			int 					factor ;
 			HANDLE					mutexhandle ;	

               
// function prototypes
            
void 			arrange_windows			(int) ;
BOOL WINAPI 	consolecontrolhandler 	(DWORD cchtype) ;
BOOL CALLBACK 	enumwindow 				(HWND, LPARAM) ;
		
			
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@
//@	 main routine
//@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 
int main (int argc, char *argv[])
{
	WSADATA 			wsa ;
    uint32              tempu ;
    int                 status ;
	DWORD				tempd ;
	int					ch ;
    int                 sock1 ;
	int					sock1addrsize ;
	struct sockaddr_in 	sock1addr ;	    		
    int                 sock2 ;
	int					sock2addrsize ;
	struct sockaddr_in 	sock2addr ;	    		
    struct ip_mreq      mreq ;
	char				rxbuff [4096] ;
	char				inputbuff [4096] ;
    int                 x ;    
    char                vlcpath [256] ;
    char                vlcparameters [256] ;
    char                temps [256] ;
	int					rx ;
	char				*pos1 ;
	char				*pos2 ;
	int					pn ;
	char				pnumber [64] ;
	char				pvalue  [64] ;
    uint32              stops ;
    uint32              nexts ;
	

    printf ("\r\n") ;
	sprintf (infoname, "whpcviewer-%s%s:  WinterHill Multi VLC Viewer", VERSIONX, VERSIONX2) ;
	printf ("%s\r\n", infoname) ;
	printf ("\r\n") ;
   
	factor = 1 ;                                    // full size display; 2 for half size
	
// parse parameters   
    
	strncpy (infoname, argv[0], sizeof(infoname)-1) ;
	pos1 = infoname ;
	do 
	{
		pos2 = strchr (pos1, '\\') ;
		if (pos2)
		{
			pos1 = pos2 + 1 ;
		}		
	}
	while (pos2) ;
	strcpy (infoname, pos1) ;
    
	myprocessid = GetCurrentProcessId() ;
 
	SetConsoleTitle (infoname) ;
	
	mutexhandle = CreateMutexA (NULL, FALSE, "whmutex320") ;
	if (mutexhandle == 0 || GetLastError() != 0)
	{	
		printf ("******* whpcviewer-%s%s is already running *******\r\n", VERSIONX, VERSIONX2) ;
		printf ("\r\n") ;
		Sleep (3000) ;
		exit (99) ;
	}
	
    if (argc != 12)
    {
        sprintf (temps, "%s", "") ;
        sprintf (temps+strlen(temps), "Usage: whpcviewer%s IPADDRESS IPPORT IPINTERFACEADDRESS ", VERSIONX) ;
        sprintf (temps+strlen(temps), "VLCLEFT VLCTOP VLCWIDTH VLCHEIGHT\r\n") ;
        sprintf (temps+strlen(temps), "INFOLEFT INFOTOP INFOWIDTH INFOHEIGHT\r\n") ;
        printf (temps) ;
        Sleep (3000) ;
		exit (1) ;
    }

    if (SetConsoleCtrlHandler(consolecontrolhandler,TRUE) == 0)
	{
		printf ("Console handler failed to install\r\n") ;
		Sleep (3000) ;
		exit (9) ;
	}               
    
// initialise signal handler
	
	stop = 0 ;

// initialise process info 	

    ZeroMemory (&si, sizeof(si)) ;
    si.cb = sizeof(si);
    ZeroMemory (&pi, sizeof(pi)) ;
	
// parse parameters	
	
	strcpy (address, argv[1]) ;
	baseport = (uint16) atoi (argv[2]) ;
	strcpy (interfaceaddress, argv[3]) ;

	if (inet_addr(address) == INADDR_NONE)
	{
		printf ("Invalid address: %s\r\n", address) ;
		Sleep (3000) ;
		exit (2) ;
	}
	
	if (atoi(argv[2]) < 1024 || atoi(argv[2]) >= 65536)
	{
		printf ("Invalid port: %d\r\n", baseport) ;
		Sleep (3000) ;
		exit (3) ;
	}
	
	if (inet_addr(interfaceaddress) == INADDR_NONE)
	{
		printf ("Invalid address: %s\r\n", interfaceaddress) ;
		Sleep (3000) ;
		exit (2) ;
	}
	
	vlcleft    = atoi (argv[4]) ;
	vlctop     = atoi (argv[5]) ;
	vlcwidth   = atoi (argv[6]) ;
	vlcheight  = atoi (argv[7]) ;
	infoleft   = atoi (argv[8]) ;
	infotop    = atoi (argv[9]) ;
	infowidth  = atoi (argv[10]) ;
	infoheight = atoi (argv[11]) ;

// look for VLC in likely places    
    
	strcpy (vlcpath, "c:\\Program Files\\VideoLAN\\VLC\\vlc.exe") ;
	if (GetFileAttributes(vlcpath) == 0xffffffff)
	{
		strcpy (vlcpath, "c:\\Program Files (x86)\\VideoLAN\\VLC\\vlc.exe") ;
		if (GetFileAttributes(vlcpath) == 0xffffffff)
		{
			strcpy (vlcpath, "\\Program Files\\VideoLAN\\VLC\\vlc.exe") ;
			if (GetFileAttributes(vlcpath) == 0xffffffff)
			{
				strcpy (vlcpath, "\\Program Files (x86)\\VideoLAN\\VLC\\vlc.exe") ;
				if (GetFileAttributes(vlcpath) == 0xffffffff)
				{
					printf ("Cannot find VLC\r\n") ;
					Sleep (3000) ;
					exit (10) ;
				}			
			}
		}
	}
	
// create socket system	
	
	if (WSAStartup(MAKEWORD(2,2),&wsa) != 0)
	{
		printf ("Socket init failed. Error Code: %d\r\n", WSAGetLastError()) ;
		Sleep (3000) ;
		exit (4) ;
	}
		       		
 	status = socket (AF_INET,SOCK_DGRAM,IPPROTO_UDP) ; 
    if (status < 0) 
    {
        printf ("Cannot create rx socket for port %d\r\n", PORTINFOLMEX) ;
		Sleep (3000) ;
        exit (1) ;
    }
    else
    {
		sock1 = status ;
    }
		       		
 	status = socket (AF_INET,SOCK_DGRAM,IPPROTO_UDP) ; 
    if (status < 0) 
    {
        printf ("Cannot create rx socket for port %d\r\n", PORTINFOMULTI) ;
		Sleep (3000) ;
        exit (1) ;
    }
    else
    {
		sock2 = status ;
    }
   
// initialise sockets   
   
    sock1addr.sin_family 			= AF_INET ; 
    sock1addr.sin_port           	= htons ((uint16)(baseport+PORTINFOLMEX)) ;
   	sock1addr.sin_addr.s_addr		= inet_addr(interfaceaddress) ; 
    memset ((void*)&sock1addr.sin_zero,0,sizeof(sock1addr.sin_zero)) ;         
    status = 1 ; 	
    status = setsockopt (sock1, SOL_SOCKET, SO_REUSEADDR, (char*)&status, sizeof(status)) ;
    if (status < 0)
    {
        printf ("Cannot re-use rx baseport\r\n") ;		
        return (status) ;
    }	
    status = bind (sock1,(const struct sockaddr*)&sock1addr,sizeof(sock1addr)) ;    
	sock1addrsize = sizeof (sock1addr) ;
    tempd = 1 ;
    status = ioctlsocket (sock1, FIONBIO, &tempd) ;		// don't wait when receiving
   
    sock2addr.sin_family 			= AF_INET ; 
    sock2addr.sin_port           	= htons ((uint16)(baseport+PORTINFOMULTI)) ;
   	sock2addr.sin_addr.s_addr		= inet_addr(interfaceaddress) ; 
    memset ((void*)&sock2addr.sin_zero,0,sizeof(sock2addr.sin_zero)) ;         
    status = 1 ; 	
    status = setsockopt (sock2, SOL_SOCKET, SO_REUSEADDR, (char*)&status, sizeof(status)) ;
    if (status < 0)
    {
        printf ("Cannot re-use rx baseport\r\n") ;		
        return (status) ;
    }	
    status = bind (sock2,(const struct sockaddr*)&sock2addr,sizeof(sock2addr)) ;     
	sock2addrsize = sizeof (sock2addr) ;
   	tempd = 1 ;
    status = ioctlsocket (sock2, FIONBIO, &tempd) ;		// don't wait when receiving

    tempu = atoi (address) ;                                // get the first octet value
    if (tempu >= 224 && tempu < 240)                        // multicast
    {    
        mreq.imr_interface.s_addr = inet_addr (interfaceaddress) ;
        mreq.imr_multiaddr.s_addr = inet_addr (address) ;

        status = setsockopt (sock1,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char*)&mreq,sizeof(mreq)) ;
        if (status < 0) 
        {
            printf ("Cannot add membership to %s for rx port %d \r\n",interfaceaddress, PORTINFOLMEX) ; 
			Sleep (3000) ;
            exit (2) ;
        }

        mreq.imr_interface.s_addr = inet_addr (interfaceaddress) ;
        mreq.imr_multiaddr.s_addr = inet_addr (address) ;

        status = setsockopt (sock2,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char*)&mreq,sizeof(mreq)) ;
        if (status < 0) 
        {
            printf ("Cannot add membership to %s for rx port %d \r\n",interfaceaddress, PORTINFOMULTI) ; 
			Sleep (3000) ;
            exit (2) ;
        }
   	}
	if (atoi(address) == 0)
	{
		strcpy (address, "") ;
	}

// Start the child processes

	for (x = 1 ; x <= MAXRECEIVERS ; x++)
	{    
		sprintf 
		(
			vlcparameters,
			" --qt-minimal-view --no-qt-video-autoresize --no-video-title-show --no-qt-system-tray udp://@%s:%d", 
			address,
			baseport + 40 + x
		) ;
		if 
		(
			!CreateProcess
			(
				vlcpath,          
				vlcparameters,
				NULL,           // Process handle not inheritable
				NULL,           // Thread handle not inheritable
				FALSE,          // Set handle inheritance to FALSE
				0,              // No creation flags
				NULL,           // Use parent's environment block
				NULL,           // Use parent's starting directory 
				&si,            // Pointer to STARTUPINFO structure
				&pi[x]          // Pointer to PROCESS_INFORMATION structure
			) 
		) 
		{
			printf ("CreateProcess failed (%d).\n", GetLastError()) ;
			Sleep (3000) ;
			exit (5) ;
		}
	}


// look through all processes to get VLC and info window handles

    do
    {
        Sleep (100) ;
        memset (vlcwindowhandle, 0, sizeof(vlcwindowhandle)) ;
        children = 0 ;
        EnumChildWindows (0, enumwindow, 7) ;
    } while (children != 4) ;
 	
	myprocessid = GetCurrentProcessId() ;
	arrange_windows (factor) ;								// start with window sizes in the command
	
// clear data and print hints
 
	strcpy (inputbuff, "") ;
	memset (rcv, 0, sizeof(rcv)) ; 
	rx = 0 ; 

	printf ("Hit ENTER or CTRL-C to quit\r\n") ;
	printf ("Hit Y to toggle half size\r\n") ;
	printf ("Hit H to hide\r\n") ;
	printf ("Hit V to display\r\n") ;
    printf ("\r\n") ;
  
// main loop  
  
	while (stop == 0) 
	{
		Sleep (100) ;												// 100ms
		if (kbhit())
		{
			ch = toupper(getch()) ;
			if (ch == LF)
			{
			}				
			else if (ch == 'Y')
			{
				factor ^= 3 ;								        // change window size factor
				arrange_windows (factor) ;
			}
			else if (ch == 'H')                                     // hide all windows
			{
				for (x = 1 ; x <= MAXRECEIVERS ; x++)
				{
					ShowWindow (vlcwindowhandle[x], SW_MINIMIZE) ;
				}
				ShowWindow (infowindowhandle, SW_MINIMIZE) ;
			}
			else if (ch == 'V')                                     // display all windows
			{
                arrange_windows (factor) ;
			}
			else if (ch == CR)                                      // quit
			{
				stop = 1 ;
			}
		}

// read the 4 line status stream        
        
		status = recvfrom (sock2, rxbuff, sizeof(rxbuff)-1, 0, (struct sockaddr*)&sock2addr, &sock2addrsize) ;
		if (status > 0)
		{
			rxbuff [status] = 0 ;
			printf (rxbuff) ;                                       // print the 4 line status stream
		}

// read the receiver info stream        
        
		status = recvfrom (sock1, rxbuff, sizeof(rxbuff)-1, 0, (struct sockaddr*)&sock1addr, &sock1addrsize) ;
		if (status > 0)
		{
			rxbuff [status] = 0 ;
			strcat (inputbuff, rxbuff) ;
			pos1 = inputbuff ;
			
			while (1)
			{
				memset (pnumber, 0, sizeof(pnumber)) ;
				memset (pvalue, 0, sizeof(pvalue)) ;
				pos2 = strchr (pos1, CR) ;
				if (pos2 == 0)
				{
					strcpy (inputbuff, pos1) ;
					break ;
				}
				pos1 = strchr (pos1, '$') ;
				if (pos1 == 0)
				{
					strcpy (inputbuff, pos1) ;
					break ;
				}
				else
				{
					pos2 = strchr (pos1 + 1, ',') ;
					if (pos2)
					{
						strncpy (pnumber, pos1 + 1, pos2 - pos1 - 1) ;
						pos1 = pos2 ;
						pos2 = strchr (pos1 + 1, CR) ;
						if (pos2)
						{
							strncpy (pvalue, pos1 + 1, pos2 - pos1 - 1) ;
							pn = atoi (pnumber) ;
							if (pn == 0 && (atoi(pvalue) <= MAXRECEIVERS))
							{
								if (rx && rcv[rx].active)
								{	
									stops = atoi (rcv[rx].textinfos[STATUS_VLCSTOPS]) ;
									nexts = atoi (rcv[rx].textinfos[STATUS_VLCNEXTS]) ;

									if (stops != rcv[rx].lastvlcstops)              // stop VLC to clear screen
									{
										SendMessage (vlcwindowhandle[rx], WM_KEYDOWN, 'S', 0) ;
										SendMessage (vlcwindowhandle[rx], WM_KEYUP,   'S', 0) ;									
										memset (rcv[rx].textinfos, 0, sizeof(rcv[rx].textinfos)) ;
                                        rcv[rx].lastvlcstops = stops ;               
									}
                                    
									if (nexts != rcv[rx].lastvlcnexts)              // send NEXT command to restart VLC
									{
										SendMessage (vlcwindowhandle[rx], WM_KEYDOWN, 'N', 0) ;
										SendMessage (vlcwindowhandle[rx], WM_KEYUP,   'N', 0) ;									
                                        rcv[rx].lastvlcnexts = nexts ;
									}
                                    
									SetWindowTextA (vlcwindowhandle[rx], rcv[rx].textinfos[STATUS_TITLEBAR]) ;
								}                                                   // update the VLC title bar
								rx = atoi (pvalue) ;
								rcv[rx].active = 1 ;
							}

							if (rx && (pn < MAXINFOS))
							{
								memset (rcv[rx].textinfos[pn], 0, sizeof(rcv[rx].textinfos[pn])) ;
								strncpy (rcv[rx].textinfos[pn], pvalue, sizeof(rcv[rx].textinfos[pn]) - 1) ;
								if (pn == 99)                                       // TS diverted
								{
									sprintf (temps,"%d:  *TS diverted to another location*", rx) ;
									SetWindowTextA (vlcwindowhandle[rx], temps) ;
									memset (rcv[rx].textinfos, 0, sizeof(rcv[rx].textinfos)) ;
									rcv[rx].active = 0 ;
								}
							}							
						}
						pos1 = pos2 + 1 ;
					}
				}				
			}						
		}	
	}	

// exited from main loop    
    
// drop multicast membership
	
    tempu = atoi (address) ;                                // get the first octet value
    if (tempu >= 224 && tempu < 240)                        // multicast
	{
		status = setsockopt (sock1,IPPROTO_IP,IP_DROP_MEMBERSHIP,(char*)&mreq,sizeof(mreq)) ;
		if (status < 0) 
		{
			printf ("Cannot drop membership to %s for rx port %d \r\n",interfaceaddress, baseport) ; 
			Sleep (3000) ;
			exit (2) ;
		}
	}
	close (sock1) ;

// Close process and thread handles and ternminate processes

    for (x = 1 ; x <= MAXRECEIVERS ; x++)
    {
        TerminateProcess (pi[x].hProcess, 0) ;
        CloseHandle (pi[x].hProcess) ;
        CloseHandle (pi[x].hThread) ;
    }    
	
	ReleaseMutex (mutexhandle) ;

	printf ("Exiting\r\n") ;
    Sleep (3000) ;                                                  // 2 second delay
    return (0) ;
}

		
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@
//@	 other routines
//@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 
// callback routine to discover window handles            
            
BOOL CALLBACK enumwindow (HWND whandle, LPARAM param)
{
    int         x ;
    DWORD       processid ;
    char        buff [256] ;
    
    (void) param ;
    
    GetWindowTextA (whandle, buff, sizeof(buff)) ;
    GetWindowThreadProcessId (whandle, &processid) ;
	if (strstr(buff, "VLC media player"))
	{
		for (x = 1 ; x <= MAXRECEIVERS ; x++)
		{
			if ((pi[x].dwProcessId == processid) && (vlcwindowhandle[x] == 0))
			{
				vlcwindowhandle[x] = whandle ;      
				children++ ;
			}
		}
	}
	
	if (strstr(buff, infoname))
	{
		infowindowhandle = whandle ;
	}
	
    return (TRUE) ;
}    


// signal handler

BOOL WINAPI consolecontrolhandler (DWORD cchtype)
{
	(void) cchtype ;
	stop = 1 ;
    Sleep (1000) ;
	return (TRUE) ;
} ;

		
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@
//@	 arrange_windows
//@
//@	 position and size the windows
//@
//@	 Calling:	size factor (divide by)
//@
//@	 Return:	
//@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 
void arrange_windows (int factor) 
{ 
    char                temps [256] ;
    int                 x ;    
	char				*pos1 ;

// first minimise the VLC windows
    
    for (x = 1 ; x <= MAXRECEIVERS ; x++)
    {
        ShowWindow (vlcwindowhandle[x], SW_MINIMIZE) ;
    }

// position the info window    
    
	pos1 = strstr (infoname, ".exe") ;
	if (pos1)
	{
		*pos1 = 0 ;
	}
	sprintf (temps, "%s : WinterHill Multi VLC Viewer", infoname) ;
    SetWindowTextA (infowindowhandle, temps) ;
    MoveWindow (infowindowhandle, infoleft + vlcleft, infotop + vlctop + vlcheight * 2 / factor, infowidth / factor, infoheight, 1) ;

    Sleep (100) ;

// restore the VLC windows    
    
    for (x = 1 ; x <= MAXRECEIVERS ; x++)
    {
        ShowWindow (vlcwindowhandle[x], SW_RESTORE) ;
    }

// position the VLC windows
    
    MoveWindow (vlcwindowhandle[1], vlcleft, vlctop, vlcwidth / factor, vlcheight / factor, 1) ;
    MoveWindow (vlcwindowhandle[2], vlcleft + vlcwidth / factor, vlctop, vlcwidth / factor, vlcheight / factor, 1) ;
    MoveWindow (vlcwindowhandle[3], vlcleft, vlctop + vlcheight / factor, vlcwidth / factor, vlcheight / factor, 1) ;
    MoveWindow (vlcwindowhandle[4], vlcleft + vlcwidth / factor, vlctop + vlcheight / factor, vlcwidth / factor, vlcheight / factor, 1) ;

// put the default text into the VLC window title bars    
    
    for (x = 1 ; x <= MAXRECEIVERS ; x++)
    {   
        sprintf (temps, "%d: udp://@%s:%d  WinterHill", x, address, baseport + 40 + x) ;
        SetWindowTextA (vlcwindowhandle[x], temps) ;
    }
}
	
